with Text_IO;                          use Text_IO;
with Pallada_Debugging;                use Pallada_Debugging;
with System;                           use System;
with Machine_Code;                     use Machine_Code;
with Line_Screen_IO, Coding;

procedure Exceptions_test is

  type Error_Type is ( None, Constraint, Numeric, Program, Machine, Wrong );

  Expected_Error : Error_Type;
  Raised_Error : Error_Type;

  procedure Test ( N : Integer ) is
  begin
    begin

      case N is

        when 1 =>
          Put_Line ( "   I range 1 .. 100 := 0" );
          Expected_Error := Constraint;
          declare
            I : Integer range 1 .. 100;
          begin
            I := 0;
          end;

        when 2 =>
          Put_Line ( "  SI range - 100 .. - 1 := 0" );
          Expected_Error := Constraint;
          declare
            Si : Short_Integer range - 100 .. - 1;
          begin
            Si := 0;
          end;

        when 3 =>
          Put_Line ( "   I := Integer'Last + 1" );
          Expected_Error := Numeric;
          declare
            I : Integer := 1;
          begin
            I := Integer'Last + I;
          end;

        when 4 =>
          Put_Line ( "  SI := Short_Integer'First - 1" );
          Expected_Error := Numeric;
          declare
            Si : Short_Integer := 1;
          begin
            Si := Short_Integer'First - Si;
          end;

        when 5 =>
          Put_Line ( "   I := Integer'Last * 2" );
          Expected_Error := Numeric;
          declare
            I : Integer := 2;
          begin
            I := Integer'Last * I;
          end;

        when 6 =>
          Put_Line ( "  SI := Short_Integer'First * 2" );
          Expected_Error := Numeric;
          declare
            Si : Short_Integer := 2;
          begin
            Si := Short_Integer'First * Si;
          end;

        when 7 =>
          Put_Line ( "   I := 1 / 0" );
          Expected_Error := Numeric;
          declare
            I : Integer := 0;
          begin
            I := 1 / I;
          end;

        when 8 =>
          Put_Line ( "  SI := ( - 1 ) / 0" );
          Expected_Error := Numeric;
          declare
            Si : Short_Integer := 0;
          begin
            Si := ( - 1 ) / Si;
          end;

        when 9 =>
          Put_Line ( "   F range 1.0 .. 100.0 := 0.0" );
          Expected_Error := Constraint;
          declare
            F : Float range 1.0 .. 100.0;
          begin
            F := 0.0;
          end;

        when 10 =>
          Put_Line ( "  SF range - 100.0 .. - 1.0 := 0.0" );
          Expected_Error := Constraint;
          declare
            Sf : Short_Float range - 100.0 .. - 1.0;
          begin
            Sf := 0.0;
          end;

        when 11 =>
          Put_Line ( "   F := Float'Last + Float'Last" );
          Expected_Error := Numeric;
          declare
            F : Float := Float'Last;
          begin
            F := F + F;
            F := F + F;
          end;

        when 12 =>
          Put_Line ( "  SF := Short_Float'First + Short_Float'First" );
          Expected_Error := Numeric;
          declare
            Sf : Short_Float := Short_Float'First;
          begin
            Sf := Sf + Sf;
            Sf := Sf + Sf;
          end;

        when 13 =>
          Put_Line ( "   F := Float'Last * 2.0" );
          Expected_Error := Numeric;
          declare
            F : Float := Float'Last;
          begin
            F := F * 2.0;
            F := F * 2.0;
          end;

        when 14 =>
          Put_Line ( "  SF := Short_Float'First * 2.0" );
          Expected_Error := Numeric;
          declare
            Sf : Short_Float := Short_Float'First;
          begin
            Sf := Sf * 2.0;
            Sf := Sf * 2.0;
          end;

        when 15 =>
          Put_Line ( "   F := 1.0 / 0.0" );
          Expected_Error := Numeric;
          declare
            F : Float := 0.0;
          begin
            F := 1.0 / F;
          end;

        when 16 =>
          Put_Line ( "  SF := - 1.0 / 0.0" );
          Expected_Error := Numeric;
          declare
            Sf : Short_Float := 0.0;
          begin
            Sf := - 1.0 / Sf;
          end;

        when 17 =>
          Put_Line ( "  1 ** ( - 1 )" );
          Expected_Error := Constraint;
          declare
            I : Integer := - 1;
          begin
            I := 1 ** I;
          end;

        when 18 =>
          Put_Line ( "  2 ** 32" );
          Expected_Error := Numeric;
          declare
            I : Integer := 32;
          begin
            I := 2 ** I;
          end;

        when 19 =>
          Put_Line ( "  2.0 ** Float'Machine_Emax" );
          Expected_Error := Numeric;
          declare
            I : Integer := Float'Machine_Emax;
            F : Float := 2.0;
          begin
            F := F ** I;
          end;

        when 21 =>
          Put_Line ( "  A ( 1 .. 10 ) ( 0 )" );
          Expected_Error := Constraint;
          declare
            A : array ( 1 .. 10 ) of Integer;
          begin
            A ( 1 ) := A ( 0 );
          end;

        when 22 =>
          Put_Line ( "  A ( 1 .. 10, 1 .. 2 ) ( 5, 0 )" );
          Expected_Error := Constraint;
          declare
            A : array ( 1 .. 10, 1 .. 2 ) of Integer;
          begin
            A ( 1, 1 ) := A ( 5, 0 );
          end;

        when 23 =>
          Put_Line ( "  A ( 1 .. 10 ) ( 0 .. 5 )" );
          Expected_Error := Constraint;
          declare
            A : array ( 1 .. 10 ) of Integer;
          begin
            A ( 1 .. 6 ) := A ( 0 .. 5 );
          end;

        when 24 =>
          Put_Line ( "  A ( 1 .. 5 ) := A ( 2 .. 7 )" );
          Expected_Error := Constraint;
          declare
            A : array ( 1 .. 10 ) of Integer;
          begin
            A ( 1 .. 5 ) := A ( 2 .. 7 );
          end;

        when 25 =>
          Put_Line ( "  P ( A ( 1 .. 2 ) => A ( 2 .. 3 ) )" );
          Expected_Error := Constraint;
          declare
            subtype S1 is String ( 1 .. 2 );
            subtype S2 is String ( 2 .. 3 );
            A : S1;
            procedure P ( A : S2 ) is
            begin
              null;
            end;
          begin
            P ( A );
          end;

        when 26 =>
          Put_Line ( "  A ( 2 .. 2 ) & A ( 1 .. 1 )" );
          Expected_Error := Constraint;
          declare
            type T is array ( 1 .. 2 ) of Integer;
            A : T;
          begin
            A := A ( 2 .. 2 ) & A ( 1 .. 1 );
          end;

        when 31 =>
          Put_Line ( "  R.C" );
          Expected_Error := Constraint;
          declare
            type R ( D : Integer ) is
              record
                case D is
                  when 1 =>
                    C : Integer;
                  when others =>
                    null;
                end case;
              end record;
            X : R ( 0 );
          begin
            X.C := 0;
          end;

        when 32 =>
          Put_Line ( "  P ( R ( 1 ) => R ( 2 ) )" );
          Expected_Error := Constraint;
          declare
            type R ( D : Integer ) is
              record
                null;
              end record;
            subtype S1 is R ( 1 );
            subtype S2 is R ( 2 );
            X : S1;
            procedure P ( X : S2 ) is
            begin
              null;
            end;
          begin
            P ( X );
          end;

        when 33 =>
          Put_Line ( "  null.all" );
          Expected_Error := Constraint;
          declare
            type A is access Integer;
            I : Integer;
            P : A;
          begin
            I := P.all;
          end;

        when 41 =>
          Put_Line ( "  Boolean'Succ ( True )" );
          Expected_Error := Constraint;
          declare
            B : Boolean := True;
          begin
            B := Boolean'Succ ( B );
          end;

        when 42 =>
          Put_Line ( "  Character'Pred ( Ascii.Nul )" );
          Expected_Error := Constraint;
          declare
            C : Character := Ascii.Nul;
          begin
            C := Character'Pred ( C );
          end;

        when 43 =>
          Put_Line ( %  Boolean'Value ( "Truue" )% );
          Expected_Error := Constraint;
          declare
            B : Boolean := Boolean'Value ( "Truue" );
          begin
            null;
          end;

        when 44 =>
          Put_Line ( "  Function without return" );
          Expected_Error := Program;
          declare
            I : Integer;
            function F return Integer is
            begin
              if    2 * 2 = 5 then
                return 0;
              end if;
            end F;
          begin
            I := F;
          end;

        when 51 =>
          Put_Line ( "  General protection (Int_13)" );
          Expected_Error := Machine;
          declare
          begin
            Mov ( Ebx, Imm ( 16#7F_FF_FF_FF# ) );
            Mov ( Mem ( Ebx ), Eax );
          end;

        when 52 =>
          Put_Line ( "  Invalid opcode (Int_6)" );
          Expected_Error := Machine;
          declare
          begin
            Db ( 16#F1# );
          end;

        when others =>
          return;

      end case;

      Raised_Error := None;

    exception
      when Constraint_Error =>
        Raised_Error := Constraint;
      when Numeric_Error =>
        Raised_Error := Numeric;
      when Program_Error =>
        Raised_Error := Program;
      when Machine_Error =>
        Raised_Error := Machine;
      when others =>
        Raised_Error := Wrong;
    end;

    if    Raised_Error = None then
      Line_Screen_IO.Bell;
      Put_Line ( "*** Failed : exception not raised!" );
    elsif Raised_Error /= Expected_Error then
      Line_Screen_IO.Bell;
      Put_Line ( "*** Failed : wrong exception raised!" );
    else
      declare
        E : constant String := Error_Type'Image ( Expected_Error );
        S : String ( E'range ) := E;
      begin
        Coding.Lower ( S ( 2 .. S'Last ) );
        Put_Line ( "    " & S & "_Error OK" );
      end;
    end if;

  end Test;

begin

  New_Line;
  Put_Line ( "                Exceptions test" );
  New_Line;

  Implicit_Raise := True;

  for I in 1 .. 60 loop
    Test ( I );
  end loop;

end Exceptions_test;
