with Unchecked_Deallocation;

with Pallada;

with Text_IO;                          use Text_IO;

  pragma Elaborate ( Text_IO );

package body Fib is

  procedure Increment is new Pallada.Increment ( Integer, Integer );
  procedure Free is new Unchecked_Deallocation ( String, Fib_Number_String_Point );

  Width : Integer := 1;
  Pred_Fib_Number : Fib_Number_String_Point := new String'( "0" );

  procedure Next is
    Z : constant := Character'Pos ( '0' );
    P : String ( 1 .. Width + 1 ); -- := ( others => '0' );
    C : Integer := 0;
    D : Integer;
  begin
    P ( Pred_Fib_Number'range ) := Pred_Fib_Number.all;
    P ( Width + 1 ) := '0';
    Free ( Pred_Fib_Number );
    Pred_Fib_Number := Fib_Number;
    for I in Fib_Number'range loop
      D := Character'Pos ( P ( I ) ) +
             Character'Pos ( Fib_Number ( I ) ) - 2 * Z + C;
      C := D / 10;
      P ( I ) := Character'Val ( Z + D mod 10 );
    end loop;
    if    C /= 0 then
      Increment ( Width );
      P ( Width ) := '1';
    end if;
    Fib_Number := new String'( P ( 1 .. Width ) );
    Increment ( Number );
  end Next;

  procedure Print is
    N : constant String := Integer'Image ( Number );
  begin
    Put ( "Fib" );
    Put ( N ( 2 .. N'Last ) );
    Put ( " = " );
    for I in reverse Fib_Number'range loop
      Put ( Fib_Number ( I ) );
    end loop;
    New_Line;
  end Print;

  procedure Short_Print is
    N : constant String := Integer'Image ( Number );
    W : constant String := Integer'Image ( Width );
  begin
    Put ( "Fib" );
    Put ( N ( 2 .. N'Last ) );
    Put ( " : Width = " );
    Put ( W ( 2 .. W'Last ) );
    New_Line;
  end Short_Print;

end Fib;
with Keyboard, Fib, Pallada, System;

  pragma Elaborate ( Fib );

package body Fib_Task is

  Signal : Character := Ascii.Nul;

  task Background is
      pragma Priority ( 1 );
  end Background;

  task Main is
      pragma Priority ( 2 );
  end Main;

  task body Background is
  begin
    loop
      if    Signal /= Ascii.Nul then
        case Signal is
          when 'p' | 'P' =>
            Fib.Print;
          when 's' | 'S' =>
            Fib.Short_Print;
          when Ascii.Cr =>
            exit;
          when others =>
            null;
        end case;
        Signal := Ascii.Nul;
      end if;
      Fib.Next;
    end loop;
  end Background;

  task body Main is
    N : Integer := 0;
  begin
    loop
      if    Keyboard.Key_Pressed then
        Signal := Keyboard.Get_Key;
        exit when Signal = Ascii.Cr;
      elsif Fib.Number >= N then
        Signal := 's';
        N := N + 1000;
      end if;
      delay 0.1;
    end loop;
  end Main;

end Fib_Task;
with Fib_Task;

procedure Fib_Test is
begin
  null;
end Fib_Test;
package Fib_Task is

end Fib_Task;
package Fib is

  Number : Integer := 1;

  type Fib_Number_String_Point is access String;
  Fib_Number : Fib_Number_String_Point := new String'( "1" );

  procedure Next;
  procedure Print;
  procedure Short_Print;

end Fib;
